/*

 Ray 1.0 - Camera handling library

*/

#ifndef _LIB_CAMERA
#define _LIB_CAMERA

#include "draw24.h"
#include "viewport.h"


//  3D - Camera:

typedef struct{

 t_basis3d basis; // camera->world transformation matrix
 t_point3d op; // camera center of projection

// the following values will be expressed in projection-plane coordinates
 FTYPE x1, y1; // top-left corner, x and y
 FTYPE x2, y2; // bottom-right corner, x and y
 FTYPE w, h; // width & height
  

/* all the following variables will be used to speed-up computations: */

// the following values will be expressed in world-space coordinates
// and will be computed starting from the above values
// and they will be directly used during ray-shooting computations
 t_point3d top_left;     // top-left corner of the projection plane
 t_point3d bottom_right; // bottom-right corner of the projection plane

// the FOVs:
 FTYPE fov_x; 
 FTYPE fov_y; 
 FTYPE d; // distance of 'op' from the projection plane

 t_viewport viewport;

} t_camera;




/* Define a navigating entity through the 3D space */

typedef struct {
 t_point3d pos; // navigator's position
 t_vector3d heading; // navigator's heading direction
// t_vector3d vel; // navigator's velocity
// t_vector3d acc; // navigator's acceleration
 t_vector3d up; // navigator's up vector
 FTYPE v; // velocity
 FTYPE accel; // acceleration
 FTYPE accel_max; // acceleration
 FTYPE v_max; // navigator's maximum velocity
 t_camera *cam;
 FTYPE scale;
} t_navigator;



t_camera g_my_camera;
t_navigator g_my_nav;




/*
Initializes a camera depending on:

 . o[x,y,z] -> origin of proj.plane
 . t[x,y,z] -> target point
 . u[x,y,z] -> up vector

 in world-space dimensions:
 . w -> width of projection plane
 . h -> height of projection plane
 . d -> distance of the origin from the projection plane
*/
void init_camera(t_camera *c, // TARGET STRUCT
								 FTYPE ox, FTYPE oy, FTYPE oz, // ORIGIN
				         FTYPE tx, FTYPE ty, FTYPE tz, // TARGET
							   FTYPE ux, FTYPE uy, FTYPE uz, // UP VECTOR
							   FTYPE w, FTYPE h, FTYPE d)    // WIDTH, HEIGHT & PROJ.PLANE DISTANCE
{

t_vector3d fwd;
t_vector3d up;
t_vector3d t;


// calculate the forward direction vector (from 'o[x,y,z]' to 't[x,y,z]')
 fwd.x = tx - ox;
 fwd.y = ty - oy;
 fwd.z = tz - oz;

 up.x = ux;
 up.y = uy;
 up.z = uz;

  create_norm_b( &c->basis,  &fwd, &up );
  
	print_b( &c->basis );

// find the origin of the proj-plane
 c->d = d;

 c->basis.o.x = ox;
 c->basis.o.y = oy;
 c->basis.o.z = oz;

// sets center-of-projection coordinates
 c->op.x = ox - c->basis.w.x*d;
 c->op.y = oy - c->basis.w.y*d;
 c->op.z = oz - c->basis.w.z*d;



// Now that we have the camera matrix (basis) we can calculate
// the projection-plane (proj-plane) parameters starting from 'w','h' and 'd'

// since we want the projection of the origin centered on the proj-plane
// its normal will be the the 'forward' vector
/*
i.e.: 
  _____
 |  |  |h/2
 |- O -|
 |__|__|h/2
 w/2 w/2
*/


 c->w = w;
 c->h = h;

 c->x1 = -0.5*w;
 c->y1 =  0.5*h;
 c->x2 =  0.5*w;
 c->y2 = -0.5*h;

// find the top-left corner coordinates (in world space)
 c->top_left.x = c->basis.o.x  +  (c->x1 * c->basis.u.x)  +  (c->y1 * c->basis.v.x);
 c->top_left.y = c->basis.o.y  +  (c->x1 * c->basis.u.y)  +  (c->y1 * c->basis.v.y);
 c->top_left.z = c->basis.o.z  +  (c->x1 * c->basis.u.z)  +  (c->y1 * c->basis.v.z);

// find the bottom-right corner coordinates (in world space)
 c->bottom_right.x = c->basis.o.x  +  (c->x2 * c->basis.u.x)  +  (c->y2 * c->basis.v.x);
 c->bottom_right.y = c->basis.o.y  +  (c->x2 * c->basis.u.y)  +  (c->y2 * c->basis.v.y);
 c->bottom_right.z = c->basis.o.z  +  (c->x2 * c->basis.u.z)  +  (c->y2 * c->basis.v.z);


 c->fov_x = 2.0 * (atan(w / (2.0*d)) / M_PI) * 180.0;
 c->fov_y = 2.0 * (atan(h / (2.0*d)) / M_PI) * 180.0;


#ifdef DEBUG_LOG
printf("\nfwd: %.2f, %.2f, %.2f,\nup : %.2f, %.2f, %.2f\no  : %.2f, %.2f, %.2f\nw-h-d: %.2f, %.2f, %.2f\n",
	   fwd.x,fwd.y,fwd.z, ux,uy,uz, ox,oy,oz, w,h,d);

 print_b(&c->basis);

 printf("cam > C tl(%.2f,%.2f,%.2f) br:(%.2f,%.2f,%.2f)\n", 
	 c->top_left.x, c->top_left.y, c->top_left.z, 
	 c->bottom_right.x, c->bottom_right.y, c->bottom_right.z );

 printf("cam > FOV x:%.2f y:%.2f\n",c->fov_x, c->fov_y );
#endif


}







// projects a point on the projection plane retrieving
// the local (u,v) coordinates
bool camera_project_point( t_point3d *src, t_camera *c, FTYPE *xp, FTYPE *yp )
{
 t_vector3d *n = &c->basis.w;
 t_point3d  *o = &c->basis.o;
 FTYPE Dx,Dy,Dz,t,den;

 Dx = src->x - c->op.x;
 Dy = src->y - c->op.y;
 Dz = src->z - c->op.z;

 den = Dx*n->x + Dz*n->z + Dy*n->y;

 if( den < 1e-6 &&  den > -1e6 ) return true;

 t = ( n->x * (o->x - c->op.x) ) + 
	 ( n->y * (o->y - c->op.y) ) + 		
	 ( n->z * (o->z - c->op.z) );

if( DISCORD(t, den) ) return true;

 t /= den;
 
// compute the intersection coordinates 
// on the projection plane
 Dx = t*Dx + c->op.x;
 Dy = t*Dy + c->op.y;
 Dz = t*Dz + c->op.z;

 Dx = Dx - o->x;
 Dy = Dy - o->y;
 Dz = Dz - o->z;

 // projection over 'u'
 n = &c->basis.u;
  *xp = Dx*n->x + Dy*n->y + Dz*n->z;
 // projection over 'v'
 n = &c->basis.v;
  *yp = Dx*n->x + Dy*n->y + Dz*n->z;

return false;

}



bool camera_project_point_screen( t_point3d *p, t_camera *c, int *x, int *y )
{
 FTYPE xp, yp;
 bool no_projection;

 no_projection = camera_project_point( p, c, &xp, &yp );

 *x = (int)(( xp / c->w + .5) * (FTYPE)SCREEN_X + .5);
 *y = (int)((-yp / c->h + .5) * (FTYPE)SCREEN_Y + .5);


 return no_projection;
}



void camera_draw_point(t_camera *c, t_point3d *p)
{
int xp, yp;

 camera_project_point_screen( p, c, &xp, &yp );

 put_pixel_32( xp, yp );
}





//////////////////////////////////






void nav_accelerate(t_navigator *nav, FTYPE ratio)
{
 if( nav->accel < nav->accel_max ) nav->accel += ratio * nav->scale;
 if( nav->v < nav->v_max ) nav->v += nav->accel;
}

void nav_decelerate(t_navigator *nav, FTYPE ratio)
{
 if( nav->accel > -nav->accel_max ) nav->accel -= ratio * nav->scale;
 if( nav->v > -nav->v_max ) nav->v += nav->accel;
}

void nav_friction(t_navigator *nav, FTYPE friction)
{
 if( friction < 0.0 ) friction = 0.0; else 
 if( friction > 1.0 ) friction = 1.0;

 nav->v = nav->v * (1.0 - friction); 

 if( ABS(nav->v) < 0.001 ) nav->v = 0.0;
}

void nav_move(t_navigator *nav)
{

 nav->pos.x += nav->heading.x * nav->v*nav->scale;
 nav->pos.y += nav->heading.y * nav->v*nav->scale;
 nav->pos.z += nav->heading.z * nav->v*nav->scale;

init_camera( nav->cam, 

 nav->pos.x, nav->pos.y, nav->pos.z,
 
 nav->pos.x + nav->heading.x,
 nav->pos.y + nav->heading.y,
 nav->pos.z + nav->heading.z,

 nav->up.x, nav->up.y, nav->up.z,

 2.0, 2.0*(SCREEN_Y/(FTYPE)SCREEN_X),  1.0);

}

void nav_follow(t_navigator *nav, int scr_x, int scr_y, FTYPE speed)
{
t_vector3d t, p; 
FTYPE f;

speed *= nav->scale;

if(speed > 1.0) speed = 1.0; else
if(speed < 0.0) speed = 0.0;

p.z = 1.0;
p.x = ( scr_x - SCREEN_X/2) / (FTYPE)SCREEN_X;
p.y = (-scr_y + SCREEN_Y/2) / (FTYPE)SCREEN_Y;


norm_v(&p);

transform_v( &t, &p, &nav->cam->basis );

nav->heading.x = nav->heading.x * (1.0-speed)  +  t.x * speed; 
nav->heading.y = nav->heading.y * (1.0-speed)  +  t.y * speed; 
nav->heading.z = nav->heading.z * (1.0-speed)  +  t.z * speed; 

norm_v( &nav->heading );

}



void nav_strife(t_navigator *nav, FTYPE dir_lx, FTYPE dir_ly, FTYPE dir_lz, FTYPE speed)
{
t_vector3d dir, delta;

dir.x = dir_lx;
dir.y = dir_ly;
dir.z = dir_lz;

norm_v_k( &dir, speed * nav->scale );

transform_v( &delta, &dir, &nav->cam->basis );

add_p_v( &nav->pos, &delta );

}

void nav_scale(t_navigator *nav, FTYPE scale)
{
 nav->scale = scale;
}


void nav_init(t_navigator *nav, t_camera *bind_cam)
{
 memset( (void *)nav, 0, sizeof(t_navigator) );

 nav->cam = bind_cam;

 nav->up.y = 1.0;
 nav->heading.z = 1.0;
 nav->v_max = .5;
 nav->accel_max = .05;

 nav->scale = 1.0;

 nav->pos.x = 0;
 nav->pos.y = 1;
 nav->pos.z = -2;

 nav_move( nav );

}


#endif

