/*

  Ray 1.0 - Lights handling library

*/

#ifndef _LIB_LIGHT
#define _LIB_LIGHT




#include "colors.h"
#include "math3d.h"
#include "mathconsts.h"
#include "billboard.h"
#include "camera.h"




#define MAX_LIGHTS 128
#define DEFAULT_LINEAR    0.0
#define DEFAULT_QUADRATIC 0.08
#define DEFAULT_RADIUS    10.0
#define DEFAULT_RADIUS_FALLOFF  8.0
#define DEFAULT_SIZE   .5
#define DEFAULT_SUBDIV  8


typedef struct {
 t_light_enum type; // type of light
 t_point3d    pos;    // light position\center
 t_vector3d   dir;    // light direction
  t_vector3d   dir_neg;    // opposite light direction (for directional lights shadow casting)
 t_rgb  color;   // light rgb color
 FTYPE dimension;  // radius of the emitting area (for soft shadows)

// Spot parameters + look-up:
 FTYPE  angle; // cut-off angle in radiants (respect to 'dir') [0 - 180(OMNI)]
 FTYPE  angle_falloff; // cosine of falloff angle
   FTYPE  angle_cos; // cosine of the cut-off angle
   FTYPE  angle_falloff_cos; // cosine of falloff angle
   FTYPE  angle_tot_cos; // cosine of 'angle + angle_fall_off'
   FTYPE  angle_const; // 1.0 / ( light->angle_cos - light->angle_falloff_cos )


 FTYPE  radius; // action radius for the light
   FTYPE radius2; // action radius for the light (squared)
 FTYPE falloff; // size of distance-falloff region
   FTYPE radius_falloff;
   FTYPE radius_falloff2; // total radius of action (squared)
   FTYPE radius_const; // "1.0 / (light->radius_falloff2 - light->radius2)"

 // attenuation factors:
 FTYPE  linear; // att. = 1.0 / (1.0 + linear*dist + quadratic*dist^2)
 FTYPE  quadratic; // att. = 1.0 / (1.0 + linear*dist + quadratic*dist^2)


 FTYPE  multiplyer; // scale the light color components
 //t_target3d *target; // tracking target
 //t_target3d *origin; // tracking origin
 
 t_billboard_map billboard_map;

} t_light;






t_light *g_lights[MAX_LIGHTS];
t_light *g_cur_light = NULL; // pointer to last allocated light
int g_num_lights = 0;






t_light *light_set_pos(t_light *l, FTYPE x, FTYPE y, FTYPE z)
{
 l->pos.x = x;
 l->pos.y = y;
 l->pos.z = z;
return l;
}

t_light *light_set_dir(t_light *l, FTYPE dx, FTYPE dy, FTYPE dz)
{
 l->dir.x = dx;
 l->dir.y = dy;
 l->dir.z = dz;

  norm_v( &l->dir );

 l->dir_neg.x = -l->dir.x;
 l->dir_neg.y = -l->dir.y;
 l->dir_neg.z = -l->dir.z;
return l;
}


t_light *light_set_at(t_light *l, FTYPE tx, FTYPE ty, FTYPE tz)
{
 return light_set_dir( l,  tx - l->pos.x, ty - l->pos.y, tz - l->pos.z );
}




t_light *light_set_color(t_light *l, FTYPE r, FTYPE g, FTYPE b)
{
 set_color3f( &l->color, r, g, b );
return l;
}


t_light *light_set_angle(t_light *l, FTYPE ang, FTYPE fall_off)
{
//
// prevents from bad angle values:
//
 // clip angle between 0 and 180
 if(ang > 180.0) ang = 180.0;
  else
 if(ang < 0.0) ang = 0.0;

 // clips fall_off to '180 - ang'
 if(fall_off > 180.0 - ang) fall_off = 180.0 - ang;
  else
 if(fall_off < 0.0) fall_off = 0.0;

 l->angle = (ang * M_PI) / 180.0;
 l->angle_cos = cos( l->angle );

 l->angle_falloff = ((ang + fall_off) * M_PI) / 180.0;
 l->angle_falloff_cos = cos( l->angle + l->angle_falloff );

 l->angle_const = 1.0 / ( l->angle_cos - l->angle_falloff_cos );


return l;
}

t_light *light_set_radius(t_light *l, FTYPE radius, FTYPE falloff )
{
 radius = MAX( radius, 0.0 );
 falloff = MAX( falloff, 0.0 );

 l->radius = radius;
 l->radius2 = radius * radius;
 l->falloff = falloff;
 l->radius_falloff  = radius + falloff;
 l->radius_falloff2 = POW2( l->radius_falloff );

 l->radius_const = 1.0 / (l->radius_falloff2 - l->radius2);

return l;
}

t_light *light_set_attenuation(t_light *l, FTYPE linear, FTYPE quadratic )
{
 if(linear < 0.0) linear = -linear;
 if(quadratic < 0) quadratic = -quadratic;

 l->linear = linear;
 l->quadratic = quadratic;

return l;
}






FTYPE get_light_attenuation(t_light *l, FTYPE dist, FTYPE dist2)
{
FTYPE att;

 if(l->quadratic)
 {
   if(l->linear==0.0)
    att  =  ( 1.0 + dist2*l->quadratic );
   else
    att  =  ( 1.0 + dist*l->linear + dist2*l->quadratic );
 }
  else
   if(l->linear)
    att  = ( 1.0 + dist*l->linear );
   else att = 1.0; // no attenuation 

 return att;


//	return (l->quadratic*dist2+l->linear*dist+1.0);
}



////////////////////////










// allocates memory for a t_light structure
t_light *allocate_light()
{
t_light *l;

  if(g_num_lights >= MAX_LIGHTS ) return NULL;

   l = (t_light *)malloc( sizeof(t_light) );

  if( l == NULL ) return NULL;
//
// Initializes default common values:
//
 //light_set_attenuation( l, DEFAULT_LINEAR, DEFAULT_QUADRATIC );
 light_set_radius( l, DEFAULT_RADIUS, DEFAULT_RADIUS_FALLOFF );

 set_color3f( &l->color,  1.0, 1.0, 1.0);
 l->multiplyer = 1.0;

make_billboard_map( &l->billboard_map, DEFAULT_SIZE, DEFAULT_SUBDIV );

 g_cur_light = g_lights[ g_num_lights ] = l;

 g_num_lights++;
 return l;
}



// creates a light of a given type with default values
t_light *make_default_light(t_light_enum type)
{
t_light *l;


if((l = allocate_light()) == NULL) return NULL;

 l->type = type;

 light_set_pos( l,  0, 0, 0 );

 light_set_dir( l,  0, 0,  1 );

 light_set_angle( l, 180.0, 0.0 );

return l;
}


// Ambient light (only the color):
t_light *make_ambient_light(FTYPE r, FTYPE g, FTYPE b)
{
 t_light *l;

 if((l = make_default_light( AMBIENT )) == NULL) return NULL;

 set_color3f( &l->color,  r, g, b );

  l->quadratic = l->linear = 0.0;

return l;
}


// Directional light (only the direction):
t_light *make_directional_light(FTYPE dx, FTYPE dy, FTYPE dz)
{
 t_light *l;

 if((l = make_default_light( DIRECTIONAL )) == NULL) return NULL;

 light_set_dir( l,  dx, dy, dz );

  l->quadratic = l->linear = 0.0;

return l;
}



t_light *make_omni_light(FTYPE ox, FTYPE oy, FTYPE oz)
{
t_light *l;

 if((l = make_default_light( OMNI )) == NULL) return NULL;

 light_set_pos( l,  ox, oy, oz );

 light_set_angle( l, 180.0, 0.0 );

return l;
}



t_light *make_spot_light(FTYPE ox, FTYPE oy, FTYPE oz, // position
						 FTYPE tx, FTYPE ty, FTYPE tz, // target
						 FTYPE angle, FTYPE fall_off)  // cut-off angles
{
t_light *l;

if( (l = allocate_light()) == NULL ) return NULL;

 l->type = SPOT;

light_set_pos( l,  ox, oy, oz );

light_set_at( l,  tx, ty, tz );

light_set_angle( l,  angle, fall_off );

return l;
}


void draw_light( t_light *l, t_camera *c)
{
if( !(l->type == SPOT || l->type==DIRECTIONAL || l->type == OMNI) ) return;

t_point3d p;

 g_MyDrawColor32 = RGBF2INT(l->color.r, l->color.g, l->color.b );
 
 for(float a=0.0; a<M_PI*2.0; a+=M_PI*2.0/5.0)
 {
	 float cs, sn;
	 cs = cosf(a)*.1;
	 sn = sinf(a)*.1;
	  p.x = l->pos.x + cs;
		p.y = l->pos.y + sn;
		p.z = l->pos.z;
	 camera_draw_point( c, &p );
	  p.x = l->pos.x + cs;
		p.y = l->pos.y;
		p.z = l->pos.z + sn;
	 camera_draw_point( c, &p );
	  p.x = l->pos.x;
		p.y = l->pos.y + sn;
		p.z = l->pos.z + cs;
	 camera_draw_point( c, &p );
 }



}



#endif


