/*

  Ray 1.0 - Scene handling library

*/


#ifndef _LIB_SCENE
#define _LIB_SCENE




#include "colors.h"
#include "lights.h"
#include "materials.h"
#include "shapes.h"



#define SCENE_MAX_NUM  2
#define SCENE_MAX_SHAPES 128
#define SCENE_MAX_LIGHTS 32

/********************/

/* Scene (set of shapes\lights) */

typedef struct {
 t_shape *shapes[ SCENE_MAX_SHAPES ]; // pointers to the shapes in the scene
 int num_objects; // total number of objects in the scene

 t_light *lights[ SCENE_MAX_LIGHTS ]; // pointers to the lights in the scene
 int num_lights;  // total number of lights in the scene
 t_rgb fog_color; // scene: fog color
 FTYPE fog_start; // distance to wich fog attenuation starts
 FTYPE fog_quadratic; // quadratic fog attenuation constant

 FTYPE clip_far; // far clipping distance
} t_scene;


t_shape *g_current_shape = NULL;
int g_shape_counter = 1; // counter used to generate shape IDs
t_scene  g_my_scene;

/********************/


void scene_free(t_scene *sc, FTYPE clip_far, FTYPE fog_r, FTYPE fog_g, FTYPE fog_b)
{
 int i;

 for( i=0; i < sc->num_objects; i++ )
 {
	 free( sc->shapes[i]->shape_data ); // free memory for specific shape data
	 free( sc->shapes[i]->material );   // free memory for material data
	 free( sc->shapes[i] );        // then free the generic container shape
 }

 for( i=0; i < sc->num_lights;  i++ )  free( sc->lights[i] );

 sc->num_objects = 0;

 sc->num_lights  = 0;

 set_color3f( &sc->fog_color, fog_r, fog_g, fog_b );

 sc->clip_far = clip_far;

 sc->fog_start = clip_far / 3.0;

 sc->fog_quadratic =  1.0 / (sc->clip_far - sc->fog_start);
}


void scene_add_shape(t_scene *sc, t_shape_enum type)
{
t_shape *shape = (t_shape *)malloc( sizeof(t_shape) );

if(shape == NULL) { err_alloc(); return; }

if( sc->num_objects >= SCENE_MAX_SHAPES ) return;


switch( type )
{
 case PLANE: make_plane( shape, g_fvals[0], g_fvals[1], g_fvals[2],
							                  g_fvals[3], g_fvals[4], g_fvals[5] ); break;

 case TRIANGLE: make_triangle(shape, 
									             g_fvals[0], g_fvals[1], g_fvals[2],
							                 g_fvals[3], g_fvals[4], g_fvals[5],
															 g_fvals[6], g_fvals[7], g_fvals[8] ); break;

 case BOX: make_box( shape );  break;


 case SPHERE: make_sphere( shape );  break;

 case CYLINDER: make_cylinder( shape );  break;

 case CONE: make_cone( shape );  break;
 
 case QUADRIC: make_quadric( shape, g_fvals[0], g_fvals[1], g_fvals[2], g_fvals[3] ); break;


 case CUBIC: break;
 
 case TRILINEAR_CELL: make_trilinear_cell( shape );  break;


 case BOX4:     make_box4( shape ); break;
 
 case TORUS:    make_torus( shape, g_fvals[0], g_fvals[1] ); break;

 case TANGLE:   make_tangle( shape ); break;

 case QUARTIC:  make_quartic( shape, g_fvals, 1.0 ); break;
 


 case METABALLS:
 case CSG:  break;

 default: 
#ifdef DEBUG_LOG
	printf("Unknown shape-type add request\n");
#endif
 break;
 
}


// Default shape setup:
srand(12345678L);

shape->id = g_shape_counter++;

shape->material = make_material( DUMMY );

//if(type == PLANE || type == BOX4 || type == QUADRIC)
shape->material->reflectance = ((rand()%64001)/64000.0)*.4+.6;
// else shape->material->reflectance = 0.0;
if(type == QUADRIC) shape->material->ior = 1.4; 
   else shape->material->ior = 0.0;

   shape->material->transparency = .9;


{
 t_rgb *c = &shape->material->diffuse;

 if(rand()%2) c->r = c->g = c->b = (rand()%128) / 127.0 + .5;
  else
	color_rainbow( &c->r, &c->g, &c->b, sc->num_objects / 2.7 );

 shape->material->texture = g_my_textures[ sc->num_objects % g_num_textures ];
 shape->material->tex = TEX_IMAGE;
}

// Initialize the shape's transformation matrices to identity
 m_identity( shape->m );
 m_identity( shape->m_inv );


 g_current_shape = sc->shapes[ sc->num_objects ] = shape;

 sc->num_objects++;

}


void scene_add_light(t_scene *sc, t_light *light)
{
 if( sc->num_lights >= SCENE_MAX_LIGHTS ) return;
 
 sc->lights[ sc->num_lights ] = light;
 sc->num_lights++;
}






#endif

