/*

  Ray 1.0 - Shapes and Attribute types definitions

*/


#ifndef _LIB_SHAPES
#define _LIB_SHAPES




#include "colors.h"
#include "lights.h"
#include "materials.h"




typedef struct {
 int x_min, y_min;
 int x_max, y_max;
 bool visible;
 bool covers_all;
} t_clipping_rect;





//=== LINEAR FUNCTIONS: =====================

/* Plane */
typedef struct {
 t_vector3d n; // plane normal
 t_point3d  o;  // plane origin
} t_plane;


/* Box */
typedef struct {
 t_plane face[6]; // face planes
 bool visible[6];   // face-visibility flags
} t_box;


/* A triangle */
typedef struct 
{
// int iv[3]; // vertex indices
 t_point3d v[3]; // vertices (TEMPORARY)
 t_vector3d n; // normal to the triangle's plane
 /* MEM! */
   t_vector3d vn[3]; // edge planes normals
   FTYPE d[3]; // edge planes constants
 /* /MEM! */
} t_triangle;
/*
//A model builded with triangular faces 
typedef struct
{
 t_point3d *pts[]; // array of the model's vertices
 t_vector3d *n[]; // vertex normals
 t_triangle *tri[]; // model's triangles
} t_model;
*/





//=== QUADRIC SURFACE: ==========================

/* Generic quadric: */
typedef struct {
 //int type; // type of quadric ('-1' if undefined)
 FTYPE A, B, C, D; // quadric equation coefficients:  Ax^2 + By^2 + Cz^2 = D
 FTYPE inv_sqrt_D; // square root of D, useful for normals computation
} t_quadric;




//=== CUBIC SHAPES: ========================

/* Trilinear cell */
typedef struct {
FTYPE val[8]; // array containing the field values at the eight corners (TEMPORARY)
FTYPE iso; // cell's iso-value
} t_trilinear_cell;

/* Scalar field (TODO) */
typedef struct {
 FTYPE *val; // array containing field samples
 int dim_x, dim_y, dim_z; // array dimensions
 int offset_line; // offset between lines
 int offset_plane;// offset between planes
 FTYPE (*field_function)(FTYPE,FTYPE,FTYPE); // 3D field function
 FTYPE iso_value; // required isovalue
} t_scalar_field;




//=== QUARTIC SHAPES: ==============

/* Torus */
typedef struct {
 FTYPE r; // torus - minor radius
 FTYPE R; // torus - major radius
 FTYPE r2, R2, rR2; // torus, squared values look-ups: r, R and (r+R)
} t_torus;


/* Quartic box */
typedef struct {
 FTYPE radius, radius_sqrt3; // radius & 'radius * sqrt(3.0)' [one half of the diagonal]
} t_box4;


/* Tangle shape */
typedef struct {
	FTYPE val; // tangle's iso value (default 11.5)
} t_tangle;


/* Generic quartic shape */
typedef struct {
 FTYPE coeffs_12[12]; // (a*X^n,b*Y^n,c*Z^n)coefficients from higher to lower quartic polynomial degree 
 FTYPE val;
} t_quartic;





//======================================
//======================================

/* General Shape */

typedef struct {

 t_shape_enum type; // shape-type identifier
 int id; // shape id
 void  *shape_data; // pointer to shape-specific data (void)

 t_material *material; // pointer to the shape's material

 t_matrix   m;     // 4x4 transformation matrix
 t_matrix   m_inv; // 4x4 inverse transformation matrix
  t_ray      ray_inv; // to store the ray inverted through 'm_inv'

 t_point3d o; // TEMPORARY: center of the shape (world-space)
 FTYPE radius; // radius of the sphere containing the shape, negative if none
  FTYPE radius2; // squared radius of the sphere containing the shape, negative if none
 t_point3d corner; // coordinates of the Box corner
 t_clipping_rect  clipping_rect;
 
 //t_plane **cap_planes; // capping planes array (needs 2 allocations for each new plane)
 t_plane *cap_planes[16]; // capping planes array (needs allocation for each new plane)
 int num_cap_planes; // number of capping planes to consider

} t_shape;



/* CSG object */
/*TODO*/
typedef struct {
 t_csg_enum op_id; // type of CSG operation to perform between A and B
 t_shape *A;  // first operator (pointer to 't_shape')
 t_shape *B;  // second operator (pointer to 't_shape')

} t_csg;





/************************************/
/************************************/



/* Memory allocation functions for the shapes: */

t_plane *plane_alloc()
{
	return  (t_plane *)malloc( sizeof(t_quadric) );
}

t_quadric *quadric_alloc()
{
	return  (t_quadric *)malloc( sizeof(t_quadric) );
}

t_box *box_alloc()
{
	return  (t_box *)malloc( sizeof(t_box) );
}


//-------------------------



void make_plane(t_shape *sh, 
			    FTYPE ox, FTYPE oy, FTYPE oz, 
			    FTYPE nx, FTYPE ny, FTYPE nz)
{
t_plane *p;

p = (t_plane *)malloc( sizeof(t_plane) );

if(p == NULL) return;

 p->o.x = ox;
 p->o.y = oy;
 p->o.z = oz;

 p->n.x = nx;
 p->n.y = ny;
 p->n.z = nz;

// since this normal is used for lighting calculations we normalize it
 norm_v( &p->n );

 sh->type = PLANE;

 sh->radius = -1;

 sh->shape_data = (void *)p;

#ifdef DEBUG_LOG
 printf("added shape - plane\n");
#endif
}




// adds a capping plane to a generic shape
// .o[x,y,z] -> origin of the plane
// .n[x,y,z] -> normal to the plane
void add_capping_plane(t_shape *sh, 
		       FTYPE ox, FTYPE oy, FTYPE oz, 
	               FTYPE nx, FTYPE ny, FTYPE nz)
{
t_plane *p;


 p = sh->cap_planes[ sh->num_cap_planes ] = (t_plane *)malloc(sizeof(t_plane));


 p->o.x = ox;
 p->o.y = oy;
 p->o.z = oz;

 p->n.x = nx;
 p->n.y = ny;
 p->n.z = nz;

// since this normal could be used for lighting calculations
// we normalize it
 norm_v( &p->n );

sh->num_cap_planes++;

#ifdef DEBUG_LOG
 printf("added capping plane\n");
#endif
}




/*  Shape-creation functions:  */
/*

 Note: all the primitives will be centered in the origin (0,0,0)

*/


void set_plane(t_plane *p, FTYPE ox, FTYPE oy, FTYPE oz, 
	                   FTYPE nx, FTYPE ny, FTYPE nz)
{

 p->o.x = ox;
 p->o.y = oy;
 p->o.z = oz;

 p->n.x = nx;
 p->n.y = ny;
 p->n.z = nz;

// since this normal could be used for lighting calculations
// we normalize it
 norm_v( &p->n );

#ifdef DEBUG_LOG
 printf("plane parameters modified\n");
#endif
}



void make_triangle(t_shape *sh, FTYPE x0, FTYPE y0, FTYPE z0, 
									              FTYPE x1, FTYPE y1, FTYPE z1, 
																FTYPE x2, FTYPE y2, FTYPE z2 )
{
t_triangle *t;
t_vector3d v[3];
int i;

 sh->shape_data = malloc( sizeof(t_triangle) );
 t = (t_triangle *)sh->shape_data;

// Set up vertices:
 t->v[0].x = x0; t->v[0].y = y0; t->v[0].z = z0;
 t->v[1].x = x1; t->v[1].y = y1; t->v[1].z = z1;
 t->v[2].x = x2; t->v[2].y = y2; t->v[2].z = z2;

// compute the triangle's normal
for(i=0; i<3; i++)  create_v( &t->v[i], &t->v[(i+1)%3],  &v[i]);

 prod_v( &t->n,  &v[0], &v[1] );
 norm_v( &t->n );


/* Edge planes Look-ups: (!___MEM___!) */
for(i=0; i<3; i++)
{
 prod_v( &t->vn[i], &v[i], &t->n );
 t->d[i] = prod_v_p( &t->vn[i], &t->v[i] );
}



t_point3d g;
 g.x = (x0 + x1 + x2) / 3.0;
 g.y = (y0 + y1 + y2) / 3.0;
 g.z = (z0 + z1 + z2) / 3.0;

 FTYPE max_dist = 0.0, dist;

for(i=0; i<3; i++)
{
 dist = sqrt( pow(t->v[i].x - g.x, 2) + pow(t->v[i].y - g.y, 2) + pow(t->v[i].z - g.z, 2) );
 if( dist > max_dist ) max_dist = dist;
}


 sh->type = TRIANGLE;
 sh->num_cap_planes = 0;

 sh->radius = max_dist;


 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = sh->radius;
}






// creates a box
void make_box(t_shape *sh)
{
int i, j;
FTYPE sx, sy, sz; // box dimensions: default 1.0 for all
t_box *b;
t_plane *p;

sh->num_cap_planes = 0;

sx = sy = sz = 1.0;

sh->shape_data = malloc( sizeof(t_box) );

b = (t_box *)sh->shape_data;

// smart face-normal\origin initialization :)

for(i=0, j=0; i<6; i++)
{
p = &b->face[i];
j=i;
 p->n.x = (j==0)*-1.0 + (j==2)*1.0;
 p->o.x = (j==0)*-sx + (j==2)*sx; 

 p->n.y = (j==1)*-1.0 + (j==3)*1.0;
 p->o.y = (j==1)*-sy + (j==3)*sy; 

 p->n.z = (j==4)*-1.0 + (j==5)*1.0;
 p->o.z = (j==4)*-sz + (j==5)*sz;


}


for(i=0; i<6; i++)  b->visible[i] = (i<3) || (i==5); // all faces are visible



 sh->type = BOX;

 sh->radius = SQRT(3.0);

 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = 1.0;


#ifdef DEBUG_LOG
 printf("added shape - box\n");
#endif

}



/*******************************/



t_quadric *make_quadric(t_shape *sh, FTYPE A, FTYPE B, FTYPE C, FTYPE D)
{
t_quadric *q;


 q = quadric_alloc();
 if( q == NULL )
 {
#ifdef DEBUG_LOG
 printf("'quadric_alloc()' returned 'NULL'\n");
#endif
	 return NULL;
 }

 sh->num_cap_planes = 0;
 sh->type = QUADRIC;

 sh->shape_data = (void *)q;

// sets quadric equation-flags
  q->A  =  A;
  q->B  =  B;
  q->C  =  C;
  q->D  =  D;
  if(D > 0.0) q->inv_sqrt_D = 1.0 / sqrt( D );
   else q->inv_sqrt_D = 1.0;

 if(A==0||B==0||C==0) sh->radius = 1;
  else sh->radius = SQRT(D);

#ifdef DEBUG_LOG
 printf("added shape - quadric(%.2f,%.2f,%.2f, %.2f)\n", A, B, C, D);
#endif

 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = sh->radius;

return q;
}


//------------------------------


// SPHERE
void make_sphere(t_shape *sh)
{

if ( make_quadric( sh,  1,1,1,1 ) == NULL) return;

 sh->num_cap_planes = 0;
// sh->type = SPHERE;
 sh->radius = 1.0;
}

// CYLINDER: axis along Y
void make_cylinder(t_shape *sh)
{

if ( make_quadric( sh, 1,0,1,1 ) == NULL) return;


// sh->type = CYLINDER;
 sh->num_cap_planes = 0;
 add_capping_plane( sh, 0,  0.5, 0, 0,  1, 0);

 add_capping_plane( sh, 0, -0.5, 0, 0, -1, 0);

 sh->radius = 1.0;
 sh->radius2 = POW2(sh->radius);

 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = sh->radius;
}

// CONE: axis along Y
void make_cone(t_shape *sh)
{

if( make_quadric( sh,  1, -1, 1, 0 ) == NULL) return;


 //sh->type = CONE;

 sh->num_cap_planes = 0;
 
 add_capping_plane( sh, 0, -0.5, 0, 0, -1, 0);

 add_capping_plane( sh, 0,  0.5, 0, 0,  1, 0);

 sh->radius = 1.0;
 sh->radius2 = 1.0;

 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = sh->radius;
}



//----------------------------------

void make_trilinear_cell(t_shape *sh)
{
t_trilinear_cell *c;

 c = (t_trilinear_cell *)malloc( sizeof(t_trilinear_cell) );

  if(c == NULL) return;

 sh->shape_data = (void *)c;

 for(int i=0; i<8; i++)  c->val[i] = 5 * (i==0||i==1||i==4||i==6);

 c->iso = 2;

 sh->type = TRILINEAR_CELL;
 sh->num_cap_planes = 0;
 sh->radius = -sqrt(3.0/4.0);
 sh->radius2 = 3.0/4.0;

 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = .5;
}
// Sets the values at the corners of the cell
void set_trilinear_cell(t_trilinear_cell *c, FTYPE iso,
												FTYPE v1, FTYPE v2, FTYPE v3, FTYPE v4,
												FTYPE v5, FTYPE v6, FTYPE v7, FTYPE v8 )
{
FTYPE *v = c->val;
c->iso = iso;

 *(v++) = v1;
 *(v++) = v2;
 *(v++) = v3;
 *(v++) = v4;
 *(v++) = v5;
 *(v++) = v6;
 *(v++) = v7;
 *v     = v8;
}



//--------------------------------------


void make_box4(t_shape *sh)
{
t_box4 *b;

 b = (t_box4 *)malloc( sizeof(t_box4) );

if(b == NULL) return;

 b->radius = 1.0;
 b->radius_sqrt3 = M_SQRT_3 * b->radius;

 sh->shape_data = (void *)b;

 sh->type = BOX4;

 sh->radius = b->radius_sqrt3;

 sh->num_cap_planes = 0;

 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = 1.0;
}


void make_torus(t_shape *sh, FTYPE r, FTYPE R)
{
t_torus *tor;


tor = (t_torus *)malloc( sizeof(t_torus) );

if(tor == NULL) return;

sh->shape_data = (void *)tor;


tor->r = r;
tor->R = R;
tor->rR2 = POW2( r + R );

tor->r2 = r*r;
tor->R2 = R*R;


 sh->type = TORUS;

 sh->num_cap_planes = 0;
 
 sh->radius = r + R;
 sh->radius2 = POW2( sh->radius );

 sh->corner.x = 
 sh->corner.y = sh->radius;
 sh->corner.z = r;
}



void make_tangle(t_shape *sh)
{
t_tangle *t;

 t = (t_tangle *)malloc( sizeof(t_tangle) );

if(t == NULL) return;

 t->val = 11.5;

 sh->shape_data = (void *)t;

 sh->type = TANGLE;

 sh->radius = 3.5;
 sh->radius2 = POW2(3.5);

 sh->corner.x = 
 sh->corner.y = 
 sh->corner.z = 2.2;

 sh->num_cap_planes = 0;
}



void make_quartic(t_shape *sh, FTYPE *coeffs_12, FTYPE val)
{
t_quartic *q;

 q = (t_quartic *)malloc( sizeof(t_quartic) );

if(q == NULL) return;

 for(int i=0; i<12; i++) q->coeffs_12[i] = coeffs_12[i];

 q->val = val;

 sh->shape_data = (void *)q;

 sh->type = QUARTIC;

 sh->radius = -1;

 sh->num_cap_planes = 0;
}
















void shape_translate(t_shape *sh, FTYPE tx, FTYPE ty, FTYPE tz)
{
 m_identity( sh->m );
 m_translate( sh->m, tx, ty, tz );
}

void shape_rotate(t_shape *sh, FTYPE ax, FTYPE ay, FTYPE az)
{
 m_identity( sh->m );
// m_rotate_x( sh->m, ax );
 m_rotate_y( sh->m, ay );
// m_rotate_z( sh->m, az );
}

void shape_scale(t_shape *sh, FTYPE sx, FTYPE sy, FTYPE sz)
{
 m_identity( sh->m );
 m_scale( sh->m, sx, sy, sz );
}








#endif

