/*

 Ray 1.0 - Solvers library

*/


#ifndef _LIB_SOLVERS
#define _LIB_SOLVERS




#include "math3d.h"
#include "matrix3d.h"
#include "shapes.h"
#include "roots3and4.h"

#include "lights.h"


typedef struct {

  FTYPE t; // solving value for the given intersection
  t_shape *shape_ptr; // pointer to the shape wich caused the intersection
  t_vector3d n; // normal to intersection point
	 t_point3d p;  // intersection coordinates

  t_rgb color; // color at intersection point
  FTYPE alpha; // opacity at intersection point

	int id_plus; // additional shape id (for the correct execution of the interpolation algorithm with some shapes)

} t_intersection;


// Epsilon value for comparisons
#define EPSILON_INT (8.0/1024.0)


//Max. allowed reflections + rifractions
#define MAX_REF 6
#define MAX_RAY_HITS (1 << MAX_REF) // 2^MAX_REF (at each hit a ray can double : reflected+refracted)





t_intersection  g_my_intersections[ MAX_RAY_HITS ];// list of intersections generated by a ray through the scene

t_intersection *g_my_intersections_ptr[ MAX_RAY_HITS ];// list of pointers for sorting purposes

t_intersection *g_cur_intersection = NULL; // for global storing purposes
// hint: # of intersections  =  g_cur_intersection - g_my_intersections

t_intersection *g_t_min_ptr = NULL;

FTYPE g_t1, g_t2, g_t_min=0.0;
int   g_hits=0;





// store "id plus" in the higher 16bits
void set_id_plus( t_intersection *t, int id_plus )
{
 t->id_plus = id_plus << 5;
}

int get_shape_id( t_shape *sh, t_intersection *t )
{
int id;

switch( sh->type )
{
 case BOX:
/*
 case TANGLE:
 case TORUS:
 case TRILINEAR_CELL:
*/
  id = sh->id | t->id_plus;

 break;

 default: 
	 id = sh->id; 
 break;
}

return id;

}






void reset_intersections(FTYPE max_t)
{
 g_unknown = 0;

 g_hits = 0;

 g_cur_intersection = g_my_intersections;

 g_t_min = max_t;

 g_t_min_ptr = NULL;
}




void get_quadric_normal(t_vector3d *n, t_shape *sh, FTYPE x, FTYPE y, FTYPE z)
{
FTYPE t;
t_quadric *q = (t_quadric *)sh->shape_data;

//switch( sh->type )

  if((q->A==-1 || q->B==-1 || q->C==-1) ) 
  {
    n->x = q->A  *  x;
    n->y = q->B  *  y;
    n->z = q->C  *  z;
	  //norm_v( n );
  }
   else
  {
   t = q->inv_sqrt_D;
    n->x = q->A  *  x * t;
    n->y = q->B  *  y * t;
    n->z = q->C  *  z * t;
  }

}







//===============================



void add_intersection(FTYPE t, t_shape *sh)
{
 if(g_hits >= MAX_RAY_HITS) return;

 g_cur_intersection->t = t;

 g_cur_intersection->shape_ptr = sh;

 if(t < g_t_min)  
 {
   g_t_min = t;
   g_t_min_ptr = g_cur_intersection;
 }

 g_cur_intersection++;
 g_hits++;

}


void add_intersection_plus(FTYPE t, t_shape *sh, int id_plus)
{
 if(g_hits >= MAX_RAY_HITS) return;

 g_cur_intersection->t = t;

 g_cur_intersection->shape_ptr = sh;

 if(t < g_t_min)  
 {
   g_t_min = t;
	 set_id_plus( g_cur_intersection, id_plus );
   g_t_min_ptr = g_cur_intersection;
 }

 g_cur_intersection++;
 g_hits++;

}



/*
finds the intersection with a plane 
 .origin -> o[x,y,z]
 .normal -> n[x,y,z]
*/
void solve_plane( t_ray *r, t_shape *sh )
{
FTYPE den;
t_plane *pl = (t_plane *)sh->shape_data;

den = pl->n.x*r->d.x + pl->n.y*r->d.y + pl->n.z*r->d.z;

//if( den > 0.1 || den < -0.1 )
if(den != 0.0)
 {

  g_t1 = ( pl->n.x * (pl->o.x - r->o.x) + 
           pl->n.y * (pl->o.y - r->o.y) + 
           pl->n.z * (pl->o.z - r->o.z) );
 
if( CONCORD( g_t1, den ) ) // positive result
  {
   g_t1/=den;
/* DISK (todo)
FTYPE x,y,z;
  x = g_t1*r->d.x + r->o.x;
  y = g_t1*r->d.y + r->o.y;
  z = g_t1*r->d.z + r->o.z;
  x = x - pl->o.x;
	den=x*x;
	x = y - pl->o.y;
  den+=x*x;
	x = z - pl->o.z;
  if(den + x*x > 64.0) return;
*/

   g_cur_intersection->n = pl->n;

   add_intersection(g_t1, sh);
  }

 }
}





void solve_triangle( t_ray *ray, t_shape *sh )
{
FTYPE den;
t_triangle *t = (t_triangle *)sh->shape_data;
t_vector3d *tn = &t->n;
t_point3d *v = t->v;


den = tn->x*ray->d.x + tn->y*ray->d.y + tn->z*ray->d.z;

// dot product
if( den > 0.01 || den < -0.01 ) // inscribing co-planar disc TODO
//if(den!=0.0)
{

  g_t1 = ( tn->x * (v[0].x - ray->o.x) + 
           tn->y * (v[0].y - ray->o.y) + 
           tn->z * (v[0].z - ray->o.z) );
 
if( CONCORD( g_t1, den ) )
 {
	FTYPE x, y, z;
  t_vector3d to_p;

  g_t1/=den; // result positive for sure

/* *** Inside-Triangle test: *** */
  x = g_t1*ray->d.x + ray->o.x;
  y = g_t1*ray->d.y + ray->o.y;
  z = g_t1*ray->d.z + ray->o.z;

/* 
MEMORY SAVING VERSION: (only triangle normal & vertices coordinates needed)
FTYPE d, p,q,r;
t_vector3d n;
// First edge: 
 p = v[1].x - v[0].x;
 q = v[1].y - v[0].y;
 r = v[1].z - v[0].z;

 n.x = q * tn->z - r * tn->y;
 n.y = r * tn->x - p * tn->z;
 n.z = p * tn->y - q * tn->x;

 d = n.x*v[0].x + n.y*v[0].y + n.z*v[0].z;

  // (edge plane normals are outward oriented)
  if( x*n.x + y*n.y + z*n.z  >  d ) return;

// Second edge:

 p = v[2].x - v[1].x;
 q = v[2].y - v[1].y;
 r = v[2].z - v[1].z;

 n.x = q * tn->z - r * tn->y;
 n.y = r * tn->x - p * tn->z;
 n.z = p * tn->y - q * tn->x;
 d = n.x*v[1].x + n.y*v[1].y + n.z*v[1].z;

  // (edge plane normals are outward oriented)
  if( x*n.x + y*n.y + z*n.z  >  d ) return;


// Third edge

 p = v[0].x - v[2].x;
 q = v[0].y - v[2].y;
 r = v[0].z - v[2].z;

 n.x = q * tn->z - r * tn->y;
 n.y = r * tn->x - p * tn->z;
 n.z = p * tn->y - q * tn->x;
 d = n.x*v[2].x + n.y*v[2].y + n.z*v[2].z;
  // (edge plane normals are outward oriented)
  if( x*n.x + y*n.y + z*n.z  >  d ) return;
*/

//---------------


 

/* big mem usage version: */
 if( x*t->vn[0].x + y*t->vn[0].y + z*t->vn[0].x  >  t->d[0] ) return;
 if( x*t->vn[1].x + y*t->vn[1].y + z*t->vn[1].x  >  t->d[1] ) return;
 if( x*t->vn[2].x + y*t->vn[2].y + z*t->vn[2].x  >  t->d[2] ) return;
/* end (High MEM version) */

//---------------

  g_cur_intersection->p.x = x;
	g_cur_intersection->p.y = y;
	g_cur_intersection->p.z = z;

  g_cur_intersection->n = *tn;

  add_intersection(g_t1, sh);
 }

}

}







void solve_box(t_ray *r, t_shape *sh)
{
int i;
FTYPE x,y,z,den;
t_box *b = (t_box *)sh->shape_data;
t_plane *pl;

FTYPE Dx,Dy,Dz, Ox,Oy,Oz;
{
FTYPE t, tt;

// ray parameters
Dx = r->d.x;
Dy = r->d.y;
Dz = r->d.z;

Ox = r->o.x;
Oy = r->o.y;
Oz = r->o.z;

// checks if the ray intersects the bounding sphere
// >>TODO ?<<< rearrange routines with trilinear cell corners translated by (-.5,-.5,-.5)

den = Ox*Ox + Oy*Oy + Oz*Oz - 3.0;
 if( den > 0 )
 {
   t = Dx*Ox + Dy*Oy + Dz*Oz;
   t *= t;
   if( t < den * (Dx*Dx + Dy*Dy + Dz*Dz) ) 
   {
    //add_intersection(.01, sh);
    return;
   }
 }
}




for(i=0; i<6; i++)
{
//if(!b->visible[i]) continue;

pl = &b->face[i];

den = pl->n.x*Dx + pl->n.y*Dy + pl->n.z*Dz;

 if( den > 0.001 || den < -0.001 )
 {

  g_t1 = ( pl->n.x * (pl->o.x - Ox) + 
           pl->n.y * (pl->o.y - Oy) + 
           pl->n.z * (pl->o.z - Oz) );

if( CONCORD( g_t1, den ) )
{
g_t1/=den; // ^ the result is for sure positive

// checks if the point belongs to the box's surface
  x = g_t1*Dx + Ox;
  y = g_t1*Dy + Oy;
  z = g_t1*Dz + Oz;

  if((x<=1+EPSILON_INT && x>=-1-EPSILON_INT) &&
	   (y<=1+EPSILON_INT && y>=-1-EPSILON_INT) &&
	   (z<=1+EPSILON_INT && z>=-1-EPSILON_INT) ) 
		if(CHECKER3D(x, y, z))
		//if(my_noise3d(x,y,z,0) > .0011)
		{
		
	      g_cur_intersection->n = pl->n; // stores the normal

				g_cur_intersection->p.x = x;
				g_cur_intersection->p.y = y;
				g_cur_intersection->p.z = z;

	      add_intersection_plus(g_t1, sh, i);
		}
}


 }

}



}


// find the roots of the quadratic equation:  at^2 + bt + c = 0
// returns the number of positive solutions
int solve_quadratic(FTYPE a, FTYPE b, FTYPE c)
{
FTYPE d;


// calculates -b/2
/* b = b * -0.5 */;
b=-b;

// calculates delta (without sqrt)
d = b*b - a*c;

// checks for delta's positivity
if(d < 0.0) return 0; // if negative return (0 solutions)

// now compute the square root of delta
d = SQRT(d);

if(a < 0.0) d=-d; // minor and major solution flips when (a < 0.0) !

// find the major solution
g_t2 = b + d;

// if the major solution would be negative we ignore all the rest (no positive solutions)
if( DISCORD(g_t2, a) ) return 0;

// else whe keep it... 
g_t2 /=a ;
//add_intersection( g_t2, sh );

//...and then we calculate the minor one
g_t1 = b - d;

if( CONCORD(g_t1, a) )  // if positive we count it (2 distinct solutions)
{
 g_t1 /= a;

// add_intersection(g_t1, sh);

 return 2;
}

// if the minor solution is negative we keep only the major one
g_t1 = g_t2;

return 1;
}





/* 
 Finds the solutions of the intersection between 'ray' 
 and a generic quadric " A*x^2 + B*y^2 + C*z^2 = D "

 So in general given a ray

   P = t * d[x,y,z] + o[x,y,z]

 to find 't' we have to solve:

  a*t^2 + b*t + c = 0

 where:
 
   a = A * dx^2 + B * dy^2 + C * dz^2
   b = 2.0*(A * dx*ox + B * dy*oy + C * dz*oz)
   c = A * ox^2 + B * oy^2 + C * oz^2 - D
   
*/
void solve_quadric(t_ray *ray, t_shape *sh)
{
 t_quadric *q = (t_quadric *)sh->shape_data;
 FTYPE a, b, c;
 FTYPE axis_d, axis_o;
 int sum=0;

 a = b = c = 0.0;

// find coefficients for the quadratic equation to solve

if(q->A)
 if(q->A > 0)
 {
  a  +=  POW2( ray->d.x );
  b  +=  ray->d.x * ray->o.x;
  c  +=  POW2( ray->o.x );
  sum++;
 }
  else
 {
  a  -=  POW2( ray->d.x );
  b  -=  ray->d.x * ray->o.x;
  c  -=  POW2( ray->o.x );
  sum--;
 }


if(q->B)
 if(q->B > 0)
 {
  a  +=  POW2( ray->d.y );
  b  +=  ray->d.y * ray->o.y;
  c  +=  POW2( ray->o.y );
  sum++;
 }
  else
 {
  a  -=  POW2( ray->d.y );
  b  -=  ray->d.y * ray->o.y;
  c  -=  POW2( ray->o.y );
  sum--;
 }


if(q->C)
 if(q->C > 0)
 {
  a  +=  POW2( ray->d.z );
  b  +=  ray->d.z * ray->o.z;
  c  +=  POW2( ray->o.z );
  sum++;
 }
  else
 {
  a  -=  POW2( ray->d.z );
  b  -=  ray->d.z * ray->o.z;
  c  -=  POW2( ray->o.z );
  sum--;
 }



if(a!=0)
{

// ************
 int solutions = solve_quadratic( a, b, c - q->D );
// ************

if(solutions)
{
FTYPE y;
FTYPE y1, y2; // capping extremes

// check if cone
if(sum==1)
{
  if(q->A==-1) { axis_d = ray->d.x; axis_o = ray->o.x; }
   else
  if(q->B==-1) { axis_d = ray->d.y; axis_o = ray->o.y; }
   else
		{ axis_d = ray->d.z; axis_o = ray->o.z; }
}

else

// Check if cylinder
if(sum==2)
{
  if(q->A==0) { axis_d = ray->d.x; axis_o = ray->o.x; }
   else
  if(q->B==0) { axis_d = ray->d.y; axis_o = ray->o.y; }
   else
		{ axis_d = ray->d.z; axis_o = ray->o.z; }
   
}

else

if(sum==3) // it's a sphere
{
 if(solutions) add_intersection( g_t1, sh );
 if(solutions==2) add_intersection( g_t2, sh );
}




if(sum < 3)
{
// Y - CAPPING:

if(sum==1) 
{ 
 if(q->D==0)	{ y1 = -1;   y2 = 0; }
				 else { y1 = -2 ;  y2 = 2; }
} else        { y1 = -1.0; y2 = 1.0; }



 switch( solutions )
 {
  case 0: break; // no solutions

//---------------

  case 1:  // one solution

	 y = g_t1 * axis_d + axis_o;		 
	 if(y>=y1 && y<=y2) add_intersection( g_t1, sh );

  break;

//---------------

  case 2:  // two solutions

	 y = g_t1 * axis_d + axis_o;		 
	 if(y>=y1 && y<=y2) add_intersection( g_t1, sh );

	 y = g_t2 * axis_d + axis_o;		 
	 if(y>=y1 && y<=y2) add_intersection( g_t2, sh );

  break;
//---------------
 } // END switch

}


}
}


}













//#####################################
//  Ray 1.0 - Trilinear cell solvers
//#####################################


/*
---------------
Trilinear cell:
---------------

      H_______G
     /|      /|
y   D-------C |
    | |     | |
^   | E_____|_F
| z |/      | /
|/  A_______B/
O--> x

A(0,0,0)


---------------------
MuPAD trilinear cell:
---------------------
x:=t*Dx+Ox;
y:=t*Dy+Oy;
z:=t*Dz+Oz;
eq:=a*(1-x)*(1-y)*(1-z) + b*x*(1-y)*(1-z) + c*x*y*(1-z) + d*(1-x)*y*(1-z) + e*(1-x)*(1-y)*z + f*x*(1-y)*z + g*x*y*z + h*(1-x)*y*z - ISO;
eqp:=poly(eq,[t]);
A:=nthcoeff(eqp,1);
B:=nthcoeff(eqp,2);
C:=nthcoeff(eqp,3);
Z:=tcoeff(eqp);

Mac:=generate::Macrofort:
Mac::init();
Mac::setOptimizedOption(TRUE):
Mac::openOutputFile("trilinear.f"):
Mac::genFor(["equal", [[c1, A],[c2, B],[c3, C],[c4, Z]]]):
Mac::closeOutputFile();
delete c1, c2, c3, c4;


----------------------
MuPAD trilinear cell gradient:
----------------------
eq:=a*(1-x)*(1-y)*(1-z) + b*x*(1-y)*(1-z) + c*x*y*(1-z) + d*(1-x)*y*(1-z) + e*(1-x)*(1-y)*z + f*x*(1-y)*z + g*x*y*z + h*(1-x)*y*z - ISO:

Gx:=diff(eq, x):
Gy:=diff(eq, y):
Gz:=diff(eq, z):

Mac:=generate::Macrofort:
Mac::init();
Mac::setOptimizedOption(TRUE):
Mac::openOutputFile("trilinear_G.f"):
Mac::genFor(["equal", [[gx, Gx],[gy, Gy],[gz, Gz]]]):
Mac::closeOutputFile();
delete gx, gy, gz;

// here some other look-ups 
// MuPad did'nt optimized it out... anyway it's only Nuts.
// (maybe it's my fault in setting up some other parameters, who knows...) *8)

*/

//-----------------------------------------------------------


void get_trilinear_cell_normal(t_vector3d *n, t_shape *sh, FTYPE x, FTYPE y, FTYPE z)
{
FTYPE *v = ((t_trilinear_cell *)sh->shape_data)->val;

FTYPE t1 = 1.0 - z;
FTYPE t2 = 1.0 - y;
FTYPE t3 = 1.0 - x;

 n->x = (v[6] - v[7])*y*z + (v[2]*y - v[3]*y)*t1 - ((v[4]*z + v[5]*z) - (v[0] + v[1])*t1)*t2;
 n->y = ((v[2] - v[1])*x + (v[3] - v[0])*t3)*t1 + ((v[6] - v[5])*x + (v[7] - v[4])*t3)*z;
 n->z = ((v[5] - v[1])*t2 + (v[6] - v[2])*y)*x + ((v[7] - v[3])*y + (v[4] - v[0])*t2)*t3;

//norm_v( n );
}








void solve_trilinear_cell(t_ray *ray, t_shape *sh)
{
/*
FTYPE a_b = ray->o.x*ray->d.x + ray->o.y*ray->d.y + ray->o.z*ray->d.z;
FTYPE a_a = POW2(ray->o.x) + POW2(ray->o.y) + POW2(ray->o.z);

// checks if the ray intersects the bounding sphere, if no return
if( POW2(a_b) < a_a - POW2(2.0) ) return 0;
*/


t_trilinear_cell *cell = (t_trilinear_cell *)sh->shape_data;
FTYPE Dx, Dy, Dz;
FTYPE Ox, Oy, Oz;
FTYPE A, B, C, D;
FTYPE *v;// pointer to cell values
FTYPE t, x,y,z;



// ray parameters
Dx = ray->d.x;
Dy = ray->d.y;
Dz = ray->d.z;

Ox = ray->o.x;
Oy = ray->o.y;
Oz = ray->o.z;

// checks if the ray intersects the bounding sphere
// >>TODO ?<<< rearrange routines with trilinear cell corners translated by (-.5,-.5,-.5)

x = Ox - 0.5;
y = Oy - 0.5;
z = Oz - 0.5;
/*
x = Ox;
y = Oy;
z = Oz;
*/


t = Dx*x + Dy*y + Dz*z;
t *= t;
A = x*x + y*y + z*z - sh->radius2;
if( A > 0 )
if( t < A * (Dx*Dx + Dy*Dy + Dz*Dz) ) 
{
 //add_intersection(.01, sh);
 return;
}

FTYPE a,b,c,d,e,f,g,h;
v = cell->val;
a = v[0];
b = v[1];
c = v[2];
d = v[3];
e = v[4];
f = v[5];
g = v[6];
h = v[7];


// ** Nice Optimized Constants :) (doing it by hand would have been damaging, W MuPAD :) ** 
/*
From here to the SolveCubic part there are a total of:
 72 MUL
 30 SUB  
 28 ADD 
 29 STO (approx)
*/

FTYPE t1,t2,t3,t4,t5,t6,
      t7,t8,t9,t10,t11,t12,
			t13,t14,t15,t16,t17,t18,
			t19,t20,t21,t22,t23;

 t1 = Dx*Oy;
 t2 = Dy*Ox;
 t3 = t1+t2;
 t4 = t3*Dz;
 t5 = 1-Oz;
 t6 = t5*Dx*Dy;
 t7 = 1-Oy;
 t8 = 1-Ox;
 t9 = t7*Dx*Dz;
 t10 = Dx*Oz;
 t11 = t7*Dx;
 t12 = t8*Dy;
 t14 = Dz*Ox;
 t15 = t10+t14;
 t16 = Dy*Oz;
 t17 = Dz*Oy;
 t18 = t16+t17;
 t19 = t11-t2;
 t20 = t12-t1;
 t21 = t8*Dz;
 t22 = t21-t10;
 t23 = -t12-t11;

 

B = a*(t6 - Dz*t23)+
    b*(-Dz*t19 - t6)+
    c*(t6 - t4)+
		d*(-Dz*t20 - t6)+
    e*(-Dy*t22 - t9)+
		f*(t9 - Dy*t15)+
    g*(t4 + Dx*t16)+
		h*(t8*Dy*Dz - Dx*t18);
      
 t13 = t17*Ox;
      
C = a*(t5*t23 - t7*t8*Dz) +
    b*(t5*t19 - t7*t14) +
		c*(t3*t5  - t13) +	
		d*(t5*t20 - t8*t17) +	
		e*(t7*t22 - t8*t16) +
		f*(t7*t15 - t16*Ox) +
		g*(t13    + t3*Oz) +
		h*(t8*t18 - t10*Oy);

D = ((g*Oz + c*t5)*Ox + (h*Oz + d*t5)*t8)*Oy +
		((b*t5 + f*Oz)*Ox + (a*t5 + e*Oz)*t8)*t7;
 
D -= cell->iso;

A = b+g+e+d-f-c-a-h;

//----------------------------------------


double coeff[4], s[3];
int sols;


// CASE (1) Full 3rd degree:
if( IS_NOT_ZERO( A ) )
{
 A *= Dx*Dy*Dz;

 coeff[3] = A;
 coeff[2] = B;
 coeff[1] = C;
 coeff[0] = D;

 // solves a cubic in the form 'c[3]*x^3 + c[2]*x^2 + c[1]*x + c[0] = 0'
 sols = solve_cubic_eq( coeff, s );
}
else
if( IS_NOT_ZERO( B ) )
{
 coeff[1] =	C/B;
 coeff[0]	= D/B;

 // solves a quadric in the form 'x^2 + c[1]*x + c[0] = 0'
 sols = solve_quadric_eq( coeff, s );
}
else
if( IS_NOT_ZERO( C ) )
{
 s[0] = -D / C;
 sols = 1;
}
 else sols = 0;

//
// Now process the solutions:
//
 while(sols--) 
 {
  if( s[ sols ] < 0.0 ) continue; // filter negative intersections

  t = s[ sols ];

 // Inside-Cube test: (Temporary ? ToDo...)
	  x = Dx * t + Ox;
		 if(x < 0) continue; if(x > 1) continue;
	  y = Dy * t + Oy;
		 if(y < 0) continue; if(y > 1) continue;
	  z = Dz * t + Oz;
		 if(z < 0) continue; if(z > 1) continue;

		g_cur_intersection->p.x = x;
		g_cur_intersection->p.y = y;
		g_cur_intersection->p.z = z;

 	  add_intersection( t, sh );
	}


}



//####################################










//############################################
/* Quartic 3D functions. Yeah! :) */
//############################################

// miscellaneous function to quickly recall 'solve_quartic()'
// and store solutions
int get_quartic_solutions(double kA, double A, double B, double C, double D, t_shape *sh)
{
int i, nroots;
double roots[4];

kA = 1.0 / kA;
A*=kA;
B*=kA;
C*=kA;
D*=kA;

  i = nroots = solve_quartic_eq(A,B,C,D, roots, SOLVE_ALL);

   while(nroots--)
   {
      FTYPE t = roots[ nroots ];
      if(t > 0.0)	add_intersection( t, sh );
	 }


return i;
}

//-------------------------------------


//=== TORUS: (x^2+y^2+z^2+R^2-r^2)^2 - 4*R*(x^2+y^2) =========================

void get_torus_normal( t_vector3d *n, t_shape *sh, FTYPE x, FTYPE y, FTYPE z )
{
t_torus *tor = (t_torus *)sh->shape_data;
FTYPE k;

k = 1.0 - tor->R / SQRT(x*x + y*y);
n->x = k * x;
n->y = k * y;
n->z = z;

//	norm_v( n );
}


// Intersect 'ray' with torus at origin, major radius R, minor radius r
int solve_torus(t_ray *ray, t_shape *sh, t_solver_enum solving_mode)
{
t_torus *tor = (t_torus *)sh->shape_data;

FTYPE t,t1,t2,t3,t4,t5,t6,t7,
      t8,t9,t10,t11,t12,t13,t14;

FTYPE kA, A, B, C, D;


  t2 = POW2( ray->d.x );
  t3 = POW2( ray->d.y );
  t4 = POW2( ray->d.z );

  t11 = POW2( ray->o.x );
  t12 = POW2( ray->o.y );
  t13 = POW2( ray->o.z );

  t5 = ray->d.x*ray->o.x;
  t6 = ray->d.y*ray->o.y;
  t7 = ray->d.z*ray->o.z;
   t = t8 = t5 + t6 + t7;

	 t14 = t11 + t12 + t13;

   t1 = t2 + t3 + t4;

if( t*t < ( t14 - tor->rR2 ) * t1 )
{
 return 0;
}

   kA = POW2( t1 );
      
	t8*=2.0;
      
   A = 2.0*t1*t8;
      
  t9  = tor->R2;
  t10 = tor->r2;

   t14 -= t9 + t10;

	 t9*=4.0;
  
   B = t4*t9 + 2*t1*t14 + POW2(t8);
   C = 2*t8*t14 + 2*t9 * t7;
   D = POW2(t14) - t9*(t10-t13);

  // Solve quartic...

return get_quartic_solutions(kA,A,B,C,D,sh);

}





//=== QUARTIC BOX:  X^4 + y^4 + z^4 = R =====================

void get_box4_normal(t_vector3d *n, t_shape *sh, FTYPE x, FTYPE y, FTYPE z)
{
FTYPE t;

t = x*x;
 n->x = x*t;

t = y*y;
 n->y = y*t;

t = z*z;
 n->z = z*t;

// norm_v( n );

}



int solve_box4(t_ray *ray, t_shape *sh)
{
t_box4 *box = (t_box4 *)sh->shape_data;
double roots[4];
FTYPE kA, A, B, C, D;


FTYPE t,t1, t2, t3, t4, t5, t6;
FTYPE c1, c2, c3;

t1 = POW2( ray->d.x );
t2 = POW2( ray->d.y );
t3 = POW2( ray->d.z );

t4 = POW2( ray->o.x );
t5 = POW2( ray->o.y );
t6 = POW2( ray->o.z );

c1 = ray->o.x*ray->d.x;
c2 = ray->o.y*ray->d.y;
c3 = ray->o.z*ray->d.z;

// checks if the ray intersects the bounding sphere
t = c1 + c2 + c3;
t *= t;

if( t < ( t4 + t5 + t6 - box->radius_sqrt3 ) * (t1 + t2 + t3) ) 
{
 //add_intersection(.01, sh);
 return 0;
}

 kA = POW2(t1) + POW2(t2) + POW2(t3);
 A = 4.0*(t1*c1 + t2*c2 + t3*c3);
 B = 6.0*(t1*t4 + t2*t5 + t3*t6);
 C = 4.0*(t4*c1 + t5*c2 + t6*c3);
 D = POW2(t4) + POW2(t5) + POW2(t6) - box->radius;


 return get_quartic_solutions(kA,A,B,C,D,sh);

}






void get_tangle_normal(t_vector3d *n, t_shape *sh, FTYPE x, FTYPE y, FTYPE z)
{
FTYPE t;
      t = x*x;
      n->x = x*(t-2.5);

      t = y*y;
      n->y = y*(t-2.5);

      t = z*z;
      n->z = z*(t-2.5);


// norm_v( n );
 
}



int solve_tangle( t_ray *ray, t_shape *sh )
{
FTYPE kA, A, B, C, D;

FTYPE Dx = ray->d.x; FTYPE Dy = ray->d.y; FTYPE Dz = ray->d.z;
FTYPE Ox = ray->o.x; FTYPE Oy = ray->o.y; FTYPE Oz = ray->o.z;

FTYPE t1,t2,t3,t4,t5,t6, c1,c2,c3;
FTYPE t;
//TANGLE: 

 t1 = Dx*Dx;
 t2 = Dy*Dy;
 t3 = Dz*Dz;

 t4 = Ox*Ox;
 t5 = Oy*Oy;
 t6 = Oz*Oz;

 c1 = Dx*Ox;
 c2 = Dy*Oy;
 c3 = Dz*Oz;

t = c1 + c2 + c3;
t *= t;
// checks if the ray intersects the bounding sphere
A = ( t4 + t5 + t6 - sh->radius2 );
if( A > 0.0 )
if( t < A * (t1 + t2 + t3) ) 
{
// add_intersection(.01, sh);
 return 0;
}


      kA = t1*t1 + t2*t2 + t3*t3;
      A = 4*(t1*c1 + t2*c2 + t3*c3);

      B = 6*(t1*t4 + t2*t5 + t3*t6) - 5*(t2 + t1 + t3);
      C = 4*(t4*c1 + t5*c2 + t6*c3) - 10*(c1 + c2 + c3);
      D = t4*t4 + t5*t5 + t6*t6 - 5*(t5 + t4 + t6) + ((t_tangle *)sh->shape_data)->val;


return get_quartic_solutions(kA,A,B,C,D,sh);
	
}  



//=============================


void get_quartic_normal(t_vector3d *n, t_shape *sh, FTYPE x, FTYPE y, FTYPE z)
{
FTYPE *c = ((t_quartic *)sh->shape_data)->coeffs_12;
FTYPE c1, c2, c3, c4;

 c1 = c[0]; c2 = c[3]; c3 = c[6]; c4 = c[9];
 n->x = POW2( x ) * 4.0*(x * c1 + (3.0/4.0) * c2) + 2.0*x * c3  +  c4;

c++;
 c1 = c[0]; c2 = c[3]; c3 = c[6]; c4 = c[9];
 n->y = POW2( y ) * 4.0*(y * c1 + (3.0/4.0) * c2) + 2.0*y * c3  +  c4;

c++;
 c1 = c[0]; c2 = c[3]; c3 = c[6]; c4 = c[9];
 n->z = POW2( z ) * 4.0*(z * c1 + (3.0/4.0) * c2) + 2.0*z * c3  +  c4;
 
 //norm_v( n );
}



int solve_quartic( t_ray *ray, t_shape *sh )
{
FTYPE val = ((t_quartic *)sh->shape_data)->val;
FTYPE *coeffs_12 = ((t_quartic *)sh->shape_data)->coeffs_12;

FTYPE Dx = ray->d.x; FTYPE Dy = ray->d.y; FTYPE Dz = ray->d.z;
FTYPE Ox = ray->o.x; FTYPE Oy = ray->o.y; FTYPE Oz = ray->o.z;

FTYPE t1, t2, t3, t4, t5, t6;
FTYPE c1, c2, c3;
FTYPE kA, A, B, C, D;
FTYPE C1x, C1y, C1z;
FTYPE C2x, C2y, C2z;
FTYPE C3x, C3y, C3z;
FTYPE C4x, C4y, C4z;

C1x = *coeffs_12++;  C1y = *coeffs_12++;  C1z = *coeffs_12++;
C2x = *coeffs_12++;  C2y = *coeffs_12++;  C2z = *coeffs_12++;
C3x = *coeffs_12++;  C3y = *coeffs_12++;  C3z = *coeffs_12++;
C4x = *coeffs_12++;  C4y = *coeffs_12++;  C4z = *coeffs_12;

c1 = Dx * Ox;
c2 = Dy * Oy;
c3 = Dz * Oz;

 t1 = POW2( Dx );
 t2 = POW2( Dy );
 t3 = POW2( Dz );
 
  kA = POW2(t1)*C1x + POW2( t2 )*C1y + POW2( t3 )*C1z;
  
  A = t1*Dx*C2x + t2*Dy*C2y + t3*Dz*C2z + 
      4.0*(t1*c1*C1x + t2*c2*C1y + t3*c3*C1z);
 
      
 t4 = POW2( Ox );
 t5 = POW2( Oy );
 t6 = POW2( Oz );
 
  B = t1*C3x + t2*C3y + t3*C3z + 
      6.0 * ( t1 * ( t4*C1x + .5*Ox*C2x ) + 
              t2 * ( t5*C1y + .5*Oy*C2y ) + 
              t3 * ( t6*C1z + .5*Oz*C2z )   );
   
                                 
  C = Dx*C4x + Dy*C4y + Dz*C4z + 
      2.0 * ( c1*C3x + c2*C3y + c3*C3z ) +
      4.0 * ( t4 * ( (3.0/4.0)*Dx*C2x + c1*C1x ) + 
              t5 * ( (3.0/4.0)*Dy*C2y + c2*C1y ) + 
              t6 * ( (3.0/4.0)*Dz*C2z + c3*C1z )   );

                    
  D = val + t4 * ( C3x + Ox*C2x ) +
            t5 * ( C3y + Oy*C2y ) +
            t6 * ( C3z + Oz*C2z ) +
            Ox*C4x + Oy*C4y + Oz*C4z + 
            POW2( t4 )*C1x + POW2( t5 )*C1y + POW2( t6 )*C1z;

return get_quartic_solutions(kA,A,B,C,D,sh);
      
}



//############################################



#endif



